//
//        Copyright 2010 Johan Dahlberg. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//    1. Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//
//    2. Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
//  AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
//  THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

var kernel                = require("./kernel");
var Ref                   = kernel.Ref;

var namewaitlist = {};
var refwaitlist = {};


kernel.on("register", function(name, ref) {
  var target;
  var callbacks;
  var index;

  if ((callbacks = namewaitlist[name])) {
    target = kernel.getActorByRef(ref);
    target._namedMonitors = callbacks;
    index = callbacks.length;
    while (index--) {
      callbacks[index]("AVAILABLE", ref, name);
    }
    delete namewaitlist[name];
  }
});



kernel.on("unregister", function(name, ref) {
  var target = kernel.getActorByRef(ref);
  var callbacks;
  var index;

  if ((callbacks = target._namedMonitors)) {
    namewaitlist[name] = target._namedMonitors;
    target._namedMonitors = null;
    index = callbacks.length;
    while (index--) {
      callbacks[index]("UNAVAILABLE", ref, name);
    }
  }
});



kernel.on("available", function(ref) {
  var target;
  var callbacks;
  var index;

  if ((callbacks = refwaitlist[ref.id])) {
    target = kernel.getActorByRef(ref);
    target._refMonitors = callbacks;
    index = callbacks.length;
    while (index--) {
      callbacks[index]("AVAILABLE", ref);
    }
    delete refwaitlist[ref.id];
  }
});



kernel.on("unavailable", function(ref, reason) {
  var target = kernel.getActorByRef(ref);
  var callbacks;
  var index;
  require("assert").ok(ref);
  if ((callbacks = target._refMonitors)) {
    refwaitlist[ref.id] = target._refMonitors;
    target._refMonitors = null;
    index = callbacks.length;
    while (index--) {
      callbacks[index](reason, ref);
    }
  }
});



exports.monitor = function(ref, callback) {
  var target;

  if (typeof callback !== "function") {
    throw new Error("bad argument, `callback`, expected Function");
  }

  if (typeof ref == "string") {
    if (/^[A-Za-z0-9]+$/.test(ref) == false) {
      throw new Error("bad argument, `ref`, named ref is not accepted");
    }
    if ((target = kernel.getActorByName(ref)) && target._available) {
      if (target._namedMonitors) {
        target._namedMonitors.push(callback);
      } else {
        target._namedMonitors = [callback];
      }
    } else {
      if (namewaitlist[ref]) {
        namewaitlist[ref].push(callback);
      } else {
        namewaitlist[ref] = [callback];
      }
    }
  } else if (ref instanceof Ref) {
    if (!(target = kernel.getActorByRef(ref)) && target._available) {
      if (target._refMonitors) {
        target._refMonitors.push(callback);
      } else {
        target._refMonitors = [callback];
      }
    } else {
      if (refwaitlist[ref.id]) {
        refwaitlist[ref.id].push(callback);
      } else {
        refwaitlist[ref.id] = [callback];
      }
    }
  } else {
    throw new Error("bad argument, `ref`, expected ref");
  }
};



exports.demonitor = function(ref, callback) {
  var target;
  var callbacks;
  var index;

  if (typeof callback !== "function") {
    throw new Error("bad argument, `callback`, expected Function");
  }

  if (typeof ref == "string") {
    if (/^[A-Za-z0-9]+$/.test(ref) == false) {
      throw new Error("bad argument, `ref`, named ref is not accepted");
    }
    if ((target = kernel.getActorByName(ref)) && target._available) {
      if ((callbacks = target._namedMonitors)) {
        if ((index = callbacks.indexOf(callback)) !== -1) {
          if (callbacks.length == 1) {
            target._namedMonitors = null;
          } else {
            target._namedMonitors.splice(index, 1);
          }
        }
      }
    } else {
      if ((callbacks = namewaitlist[ref])) {
        if ((index = callbacks.indexOf(callback)) !== -1) {
          if (callback.length == 1) {
            delete namewaitlist[ref];
          } else {
            callbacks.splice(index, 1);
          }
        }
      }
    }
  } else if (ref instanceof Ref) {
    if ((target = kernel.getActorByRef(ref)) && target._available) {
      if ((callbacks = target._refMonitors)) {
        if ((index = callbacks.indexOf(callback)) !== -1) {
          if (callbacks.length == 1) {
            target._refMonitors = null;
          } else {
            target._refMonitors.splice(index, 1);
          }
        }
      }
    } else {
      if ((callbacks = refwaitlist[ref.id])) {
        if ((index = callbacks.indexOf(callback)) !== -1) {
          if (callbacks.length == 1) {
            delete refwaitlist[ref.id];
          } else {
            callbacks.splice(index, 1);
          }
        }
      }
    }
  } else {
    throw new Error("bad argument, `ref`, expected ref");
  }
};